/*
 * Copyright (c) 2005-2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <portLib.h>

#include "server/jlocoHeader.h"

#include "jlocoPrivate.h"
#include "jlocoConst.h"
#include "odoProba.h"
#include "joystick.h"
#include <contlaw/contLaw.h>
#include <errno.h>
#include <math.h>
#include <pthread.h>
#include <stdlib.h>

// NOTE: if FAKE_NEOBOTIX is defined, jlocoPrivate.h declares the
// required subset of the neobotix API.
#ifndef FAKE_NEOBOTIX
# include <neobotix/neobotixErrors.h>
# include <neobotix/pscan.h>
# include <neobotix/neo.h>
# include <neobotix/canMotors.h>
# include <neobotix/canGyro.h>
# include <neobotix/canIO.h>
# define CAN_DEVICE_PATH "/dev/pcan24"
#endif // FAKE_NEOBOTIX


// like PDEBUG, but always verbose for the most important error messages
#define PERRMSG(fmt, args...) fprintf(stderr, "%s(): "fmt, __func__, ## args)


#define NEO_MOTORS_MAXNTRIALS 3


// If an INVERT symbol is defined "1", then the corresponding wheel
// speeds and increments will be inverted. If it defined "0", then
// these values will be laft as-is. Any other definition (as well as
// missing definitions) are treated as a compiler error.
#define INVERT_LEFT_WHEEL  1
#define INVERT_RIGHT_WHEEL 0


// TODO: Make joystick device path runtime-configurable
#ifndef FAKE_JOYSTICK
# define JOYSTICK_DEVICE_PATH "/dev/input/js0"
#else
# define JOYSTICK_DEVICE_PATH "/dev/zero"
#endif


//////////////////////////////////////////////////
// begin
// static function prototypes

// IMPORTANT: call this before using SDI_F->cmdParam, ie before
// initializing odometry which uses distPoint!
static void initCmdParam(GENPOS_CART_CMD_PARAM *cmd);

// IMPORTANT: call this before using SDI_F->varParams, ie before
// initializing odometry which requires (co)variances!
static void initVarParams(JLOCO_VAR_PARAMS *vp);

static void initOdoModel(JLOCO_PRIVATE *jp);

static void initLimits(JLOCO_PRIVATE *jp);

static void initWatchdogs(JLOCO_PRIVATE *jp);

static void initPomPosters();

static void updatePomPosters(JLOCO_PRIVATE *jp);

// Measure wheel increments and update the cache of actuator and
// global speed values. The odo struct gets changed in-place. Returns
// 0 on success, or the return value of receiveMotorChange() if that
// failed.
static int updateSpeedCache(JLOCO_ODO *odo, double motRr, double motRl,
			    double motW);

/* convert global (v, w) to wheel (wr, wl) speeds */
static void speedGlobToWheels(double v, double w,
			      double *wRight, double *wLeft,
			      double motRr, double motRl, double motW);

/*  compute speed commands that don't violate dynamic limits;
 *  can be used for linear and angular speeds */
static double applyLimits(double current, double wanted,
			  double limit, double acceleration,
			  double timeslice);

static void getJoystickSpeeds(double *v, double *w);

/* Verify that the robot actually moves when commands are
 * sent. Returns: 0 if the check passes, 1 if not.
 */
static int checkRobotMotion(double vWanted, double wWanted,
			    double vCurrent, double wCurrent);

/* Integrate odometry based on the speed values stored in the
 * JLOCO_ODO structure (call updateSpeedCache() before). Returns 0 on
 * success.
 */
static void updateOdometry(JLOCO_ODO *odo, double delta_gyro_angle,
			   double motW, double qCoeff);

/* Checks canMotorsState() on both motors, restarts them with
   canMotorsReset() if necessary. Returns 0 on success, fails with -1
   if canMotorsState() returned ST_MOTOR_FAILURE, with -3 if
   canMotorsReset() did not return OK, with -2 if canMotorsReset()
   succeeded but canMotorsState() then failed more than maxtrials
   times.
*/
static int keepMotorsAlive(NEO_STR *neo, int maxtrials);

/* Calls keepMotorsAlive() before transmitting the speeds via
   canMotorsSetVelGearRadS(). Returns 0 on success, or the non-zero
   return value of keepMotorsAlive().
*/
static int transmitMotorSpeeds(NEO_STR *neo, double lvel, double rvel,
			       int maxtrials);

/* Calls keepMotorsAlive() before reading the motor speeds and
   increments since last call via
   canMotorsGetGearDeltaPosVelRadS(). Returns 0 on success, or the
   non-zero return value of keepMotorsAlive().
*/
static int receiveMotorChange(NEO_STR *neo, double *lDeltaPos, double *lvel,
			      double *rDeltaPos, double *rvel, int maxtrials);

// end
// static function prototypes
//////////////////////////////////////////////////


static NEO_STR *neo = NULL;
static JOYSTICK_STR *joy = NULL;
static POSTER_ID pomPosterId = NULL;
static BOOL jlocoInitDone = FALSE;

/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_jloco_NEOBOTIX_ERROR
 *              S_jloco_MOTOR_ERROR
 *              S_jloco_JOYSTICK_ERROR
 *              S_jloco_MOTOR_READ_ERROR
 *              S_jloco_INVALID_TRACK_MODE
 *              S_jloco_BATTERY_LOW
 *              S_jloco_SIGACTION_ERROR
 */

/* jlocoInitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
jlocoInitStart(int *report)
{
  //////////////////////////////////////////////////
  // start neobotix driver
  jlocoInitDone = FALSE;
  if (neo) {
    printf ("jlocoInitStart: %s already spawned\n", NEO_READ_TASK_NAME);
    return EXEC;
  }
  neo = neoInit(CAN_DEVICE_PATH);
  SDI_F->batLevel = JLOCO_MIN_BATTERY_LEVEL;
  if(NULL == neo){
    PERRMSG("neoInit(%s) failed: %s\n",
	    CAN_DEVICE_PATH, h2getMsgErrno(errnoGet()));
    *report = S_jloco_NEOBOTIX_ERROR;
    return ETHER;
  }
  return EXEC;
}

/* jlocoInitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
jlocoInitMain(int *report)
{
  //////////////////////////////////////////////////
  // initialize neobotix drivers and joystick
  if(ERROR == canMotorsStart(neo)){
    PERRMSG("canMotorsStart() failed: %s\n", h2getMsgErrno(errnoGet()));
    *report = S_jloco_MOTOR_ERROR;
    return ETHER;
  }
  canIORequestConnect(neo);
  canGyroRequestConnect(neo);
  canGyroZero(neo);
#ifndef FAKE_JOYSTICK
  joy = joystickInit(JOYSTICK_DEVICE_PATH);
#else
  joy = (JOYSTICK_STR *)calloc(1, sizeof(JOYSTICK_STR));
#endif
  if(NULL == joy){
    PERRMSG("joystickInit(%s) failed: %s\n",
	    JOYSTICK_DEVICE_PATH, strerror(errno));
    *report = S_jloco_JOYSTICK_ERROR;
    return ETHER;
  }
  jlocoInitDone = TRUE;
  return ETHER;
}


/*------------------------------------------------------------------------
 * initOdoAndAsserv  -  Initialization codel of CmdTask
 * 
 * Returns:    OK or ERROR
 */
STATUS
initOdoAndAsserv(int *report)
{
  //////////////////////////////////////////////////
  // initialize jlocoPrivate
  jlocoPrivate->trackMode     = JLOCO_IDLE;
  jlocoPrivate->cmdTaskPeriod = EXEC_TASK_PERIOD(JLOCO_CMDTASK_NUM);
  // no need to initialize posRef
  // no need to initialize speedRef
  // no need to initialize trackedPoster
  initOdoModel(jlocoPrivate);
  initLimits(jlocoPrivate);
  initWatchdogs(jlocoPrivate);
  // before odometry, need to initialize cmdParam and varParam
  initCmdParam(&SDI_F->cmdParam);
  initVarParams(&SDI_F->varParams);
  initOdometry(jlocoPrivate, 0, 0, 0, SDI_F->varParams.vxx,
	       SDI_F->varParams.vyy, SDI_F->varParams.vtt,
	       SDI_F->cmdParam.distPoint);
  
  //////////////////////////////////////////////////
  // initialize SDI_F
  // cmdParam and varParams already initialized above
  initPomPosters();
  SDI_F->gyroMode = JLOCO_DEFAULT_GYRO_MODE;
  
#ifndef FAKE_NEOBOTIX
  neobotixRecordH2errMsgs();
#endif

  return OK;
}


/*------------------------------------------------------------------------
 * odoAndAsserv  --  permanent activity of CmdTask
 * 
 * Returns: OK or ERROR
 */
STATUS
odoAndAsserv(int *report)
{
  static double prevV = 0;	// latch vRefLim
  static double prevW = 0;	// latch wRefLim
  double vCurrent = 0;	        // will be initialized from MEASURED value
  double wCurrent = 0;	        // will be initialized from MEASURED value
  double vRefRaw = 0; 	// linear velocity command w/o dynamic limits
  double wRefRaw = 0;	// angular velocity command w/o dynamic limits
  int trajectory_transition = 0; // set to 1 if watchdog is -1

  // not always needed, but better don't clutter code with tons of
  // #ifdef directives... it only cost a couple of variable
  // assignments at runtime
  char * msg = 0;
#ifdef GETOUTSTATEDBG
  static char * prevmsg = 0;
#endif
  char * ststr = "<illegal>";
  
  /* Initialisation not done */
  if (!jlocoInitDone) return OK;

  PVDEBUG("\n\n--------------------------------------------------\n");
  
  // Request new data.
  canIORequestVBatt(neo);
  canIORequestDigiIn(neo);
  canGyroRequestAngle(neo);
  
  //////////////////////////////////////////////////
  // Emergency stop preempts all other.
  if(canIOGetDigiIn(neo) & CAN_IO_EMERGENCY_STOP){
    int res;
    if(JLOCO_EMERGENCY != jlocoPrivate->trackMode){
      PDEBUG("EMERGENCY BREAK\n");
    }
    jlocoPrivate->trackMode = JLOCO_EMERGENCY;
    msg = "EMERGENCY BREAK";
    res = transmitMotorSpeeds(neo, 0, 0, NEO_MOTORS_MAXNTRIALS);
    if(0 != res){
      PERRMSG("transmitMotorSpeeds() returned %d\n", res);
      *report = S_jloco_MOTOR_ERROR;
/*       return ERROR; */
    }
    return OK;
  }
  
  //////////////////////////////////////////////////
  // Update odometry and speed state.
  {
    static int prevmoving = 0;
    static double prevgyro = 0;
    int moving;
    double gyro;
    int res;
    res = updateSpeedCache(&jlocoPrivate->odo, jlocoPrivate->motRr_cor,
			   jlocoPrivate->motRl_cor, jlocoPrivate->motW_cor);
    if(0 != res){
      PERRMSG("updateSpeedCache() returned %d\n", res);
      *report = S_jloco_MOTOR_READ_ERROR;
      jlocoPrivate->trackMode = JLOCO_EMERGENCY;
    }
    // see if we're moving at all
    if((fabs(jlocoPrivate->odo.rightWheel.v) > JLOCO_EPSILON)
       || (fabs(jlocoPrivate->odo.leftWheel.v) > JLOCO_EPSILON)){
      moving = 1;
      vCurrent = jlocoPrivate->odo.robot.v;
      wCurrent = jlocoPrivate->odo.robot.w;
    }
    else{
      moving = 0;
      vCurrent = 0;
      wCurrent = 0;
    }
    // detect start of new movement
    if((0 == prevmoving) && (0 != moving)){
      canGyroZero(neo);
      prevgyro = 0;
      gyro = 0;
    }
    else{
      gyro = canGyroGetAngle(neo);
    }
    // integrate odometry when moving
    if(0 != moving)
      updateOdometry(&jlocoPrivate->odo, gyro - prevgyro,
		     jlocoPrivate->motW_cor, SDI_F->varParams.qCoeff);
    prevmoving = moving;
    prevgyro = gyro;
  }

  // the poster for POM
  updatePomPosters(jlocoPrivate);
  
  //////////////////////////////////////////////////
  // Check for manual override, preempted by subsequent checks.

  if (JLOCO_MANUAL == jlocoPrivate->trackMode) {

#ifdef FAKE_JOYSTICK
    PVDEBUG("FAKE_JOYSTICK ==> not polling joystick\n");
    memset(&joy->buttons, 0, JOYSTICK_N_BUTTONS);
    memset(&joy->axes,    0, JOYSTICK_N_AXES);
#else // FAKE_JOYSTICK
      int joyresult;
      PVDEBUG("polling joystick\n");
      joyresult = joystickPoll(joy);
      if(0 > joyresult){
	PERRMSG("joystickPoll() returned %d\n", joyresult);
	*report = S_jloco_JOYSTICK_ERROR;
	jlocoPrivate->trackMode = JLOCO_EMERGENCY;
      }
      PVDEBUG("joystickPoll() returned %d\n", joyresult);
#endif // FAKE_JOYSTICK
  }
    
  //////////////////////////////////////////////////
  // Check if previous motion command is respected.
  {
    static int cmdMismatchCounter = 0;
    if(JLOCO_IDLE != jlocoPrivate->trackMode){
      if(0 == checkRobotMotion(prevV, prevW, vCurrent, wCurrent))
	cmdMismatchCounter = 0;
      else
	++cmdMismatchCounter;
      if(cmdMismatchCounter > jlocoPrivate->cmdMismatchThreshold){
	PDEBUG("motion commands not respected\n");
	msg = "cmd mismatch";
	jlocoPrivate->trackMode = JLOCO_IDLE;
      }
    }
  }
  
  //////////////////////////////////////////////////
  // Check battery voltage.
  {
#ifdef DISABLE_BATCHECK
# warning Battery voltage check disabled.
    PVDEBUG("Battery voltage check disabled.\n");
#else
    SDI_F->batLevel = canIOGetVBatt(neo);
#endif
    if(SDI_F->batLevel < JLOCO_MIN_BATTERY_LEVEL){ // TODO: make config'able
      PERRMSG("battery low: %d < %d\n", SDI_F->batLevel, 
	  JLOCO_MIN_BATTERY_LEVEL);
      *report = S_jloco_BATTERY_LOW;
      return OK;
    }
  }
  
  //////////////////////////////////////////////////
  // Check drift flag (only in position track mode)
  if(JLOCO_TRACK_POSITION == jlocoPrivate->trackMode
     && jlocoPrivate->importantDrift){
      PDEBUG("important drift detected\n");
      msg = "important drift";
      jlocoPrivate->trackMode = JLOCO_IDLE;
    }
  
  //////////////////////////////////////////////////
  // Check internal watchdog if tracking a poster
  // ATTENTION a value of -1 triggers special behavior: keep mode but
  // set zero speeds (needed for transitions between trajectories)
  if(   JLOCO_TRACK_POSITION == jlocoPrivate->trackMode
     || JLOCO_TRACK_SPEED    == jlocoPrivate->trackMode){
    // WARNING: Initializing prevPumpNumRef to zero is important when
    // working with pilo.
    static int prevPumpNumRef = 0;
    static int pumpMismatchCounter = 0;
    int pumpNumRef;
    double v, w;
    
    if(JLOCO_TRACK_POSITION == jlocoPrivate->trackMode){
      v = jlocoPrivate->posRef.v;
      w = jlocoPrivate->posRef.w;
      pumpNumRef = jlocoPrivate->posRef.numRef;
    }
    else{
      v = jlocoPrivate->speedRef.v;
      w = jlocoPrivate->speedRef.w;
      pumpNumRef = jlocoPrivate->speedRef.numRef;
    }
    
    if(-1 == pumpNumRef)
      trajectory_transition = 1;
    else if(pumpNumRef != prevPumpNumRef)
      pumpMismatchCounter = 0;
    else{
      // special case: watchdog dead but zero speed reference is OK
      if(fabs(v) < JLOCO_EPSILON && fabs(w) < JLOCO_EPSILON)
	trajectory_transition = 1;
      else{
	++pumpMismatchCounter;
	PDEBUG("pump watchdog no change %d (%d)\n",
		pumpMismatchCounter, pumpNumRef);
	if(pumpMismatchCounter > jlocoPrivate->pumpMismatchThreshold){
	  PDEBUG("pump watchdog no change %d exceeds threshold %d\n",
		 pumpMismatchCounter, jlocoPrivate->pumpMismatchThreshold);
	  msg = "pump task is dead";
	  jlocoPrivate->trackMode = JLOCO_IDLE;
	}
      }
    }
    prevPumpNumRef = pumpNumRef;
  }

  SDI_F->trackMode = jlocoPrivate->trackMode;

  //////////////////////////////////////////////////
  // Compute vRefRaw and wRefRaw depending on trackMode

  switch(jlocoPrivate->trackMode){
  case JLOCO_IDLE:
    PVDEBUG("JLOCO_IDLE\n");
    ststr = "IDLE";
    vRefRaw = 0;
    wRefRaw = 0;
    break;
    
  case JLOCO_MANUAL:
    PVDEBUG("JLOCO_MANUAL\n");
    ststr = "MANUAL";
    getJoystickSpeeds(&vRefRaw, &wRefRaw);
    break;
    
  case JLOCO_TRACK_SPEED:
    if(0 == trajectory_transition){
      ststr = "TRACK_SPEED";
      PVDEBUG("JLOCO_TRACK_SPEED\n");
      vRefRaw = jlocoPrivate->speedRef.v;
      wRefRaw = jlocoPrivate->speedRef.w;
    }
    else{
      ststr = "TRACK_SPEED [trans]";
      PVDEBUG("JLOCO_TRACK_SPEED traj transition\n");
      vRefRaw = 0;
      wRefRaw = 0;
    }
    break;
    
  case JLOCO_TRACK_POSITION:
    if(0 == trajectory_transition){
      JLOCO_ODO *odo = & jlocoPrivate->odo;
      const double distPoint = SDI_F->cmdParam.distPoint;
      const double cosTheta = odo->robotMisc.cosTheta;
      const double sinTheta = odo->robotMisc.sinTheta;
      ststr = "TRACK_POS";
      PVDEBUG("JLOCO_TRACK_POSITION\n");
      // compute "controlled point" and use contlaw library for tracking
      if(GENPOS_BACKWARD_MOTION == jlocoPrivate->posRef.backFlag){
	odo->robot.xRef = odo->robot.xRob - distPoint * cosTheta;
	odo->robot.yRef = odo->robot.yRob - distPoint * sinTheta;
      }
      else{
	odo->robot.xRef = odo->robot.xRob + distPoint * cosTheta;
	odo->robot.yRef = odo->robot.yRob + distPoint * sinTheta;
      }
      jlocoPrivate->importantDrift = 
	contLawCartPositionControl(&SDI_F->cmdParam,
				   &jlocoPrivate->posRef,
				   &odo->robot,
				   jlocoPrivate->cmdTaskPeriod,
				   prevV, prevW, &vRefRaw, &wRefRaw);
    }
    else{
      ststr = "TRACK_POS [trans]";
      PVDEBUG("JLOCO_TRACK_POSITION with trajectory transition\n");
      vRefRaw = 0;
      wRefRaw = 0;
    }
    break;
    
  case JLOCO_EMERGENCY:
    ststr = "EMERGENCY";
    {
      const int res = transmitMotorSpeeds(neo, 0, 0, NEO_MOTORS_MAXNTRIALS);
      if(0 != res){
	PERRMSG("transmitMotorSpeeds() returned %d\n", res);
	*report = S_jloco_MOTOR_ERROR;
/* 	return ERROR; */
      }
    }
    jlocoPrivate->trackMode = JLOCO_IDLE;
    return OK;
    
  default:
    PERRMSG("invalid trackMode %d\n", jlocoPrivate->trackMode);
    {
      const int res = transmitMotorSpeeds(neo, 0, 0, NEO_MOTORS_MAXNTRIALS);
      if(0 != res)
	PERRMSG("AND transmitMotorSpeeds() returned %d, TOO!\n", res);
    }
    *report = S_jloco_INVALID_TRACK_MODE;
    return ERROR;
  }
  
  //////////////////////////////////////////////////
  // apply dynamic limits, transform to actuator space
  {
    double vRefLim = 0;	// linear velocity command with dynamic limits
    double wRefLim = 0;	// angular velocity command with dynamic limits
    double wLeft = 0;	// left wheel velocity command
    double wRight = 0;	// right wheel velocity command
    
    vRefLim = applyLimits(vCurrent, vRefRaw,
			  jlocoPrivate->sdMax, jlocoPrivate->sddMax, 
			  jlocoPrivate->cmdTaskPeriod);
    wRefLim = applyLimits(wCurrent, wRefRaw,
			  jlocoPrivate->thetadMax, jlocoPrivate->thetaddMax, 
			  jlocoPrivate->cmdTaskPeriod);
    speedGlobToWheels(vRefLim, wRefLim, &wRight, &wLeft,
		      jlocoPrivate->motRr_cor, jlocoPrivate->motRl_cor,
		      jlocoPrivate->motW_cor);
    
#ifdef GETOUTSTATEDBG
    // state line
    if(0 == msg){
      if(0 == prevmsg)
	fprintf(stderr, "%s\n", ststr);
      else
	fprintf(stderr, "%s (previous msg \"%s\")\n", ststr, prevmsg);
    }
    else{
      if((msg == prevmsg) || (0 == prevmsg))
	fprintf(stderr, "%s (msg \"%s\")\n", ststr, msg);
      else
	fprintf(stderr, "%s (new msg \"%s\" previous msg \"%s\")\n",
		ststr, msg, prevmsg);
    }
    // odometry line
    if(JLOCO_TRACK_POSITION == jlocoPrivate->trackMode)
      fprintf(stderr, "  pos (%5.2f   %5.2f   %5.2f)  cntrl (%5.2f   %5.2f)\n",
	      jlocoPrivate->odo.robot.xRob, jlocoPrivate->odo.robot.yRob,
	      jlocoPrivate->odo.robot.theta, jlocoPrivate->odo.robot.xRef,
	      jlocoPrivate->odo.robot.yRef);
    else
      fprintf(stderr, "  pos (%5.2f   %5.2f   %5.2f)\n",
	      jlocoPrivate->odo.robot.xRob, jlocoPrivate->odo.robot.yRob,
	      jlocoPrivate->odo.robot.theta);
    // reference line
    if(JLOCO_TRACK_POSITION == jlocoPrivate->trackMode)
      fprintf(stderr, "  ref (%5.2f   %5.2f   %5.2f)  speed (%5.2f   %5.2f)\n",
	      jlocoPrivate->posRef.x, jlocoPrivate->posRef.y,
	      jlocoPrivate->posRef.theta, jlocoPrivate->posRef.v,
	      jlocoPrivate->posRef.w);
    else if(JLOCO_TRACK_SPEED == jlocoPrivate->trackMode)
      fprintf(stderr, "  ref (%5.2f   %5.2f)\n",
	      jlocoPrivate->speedRef.v, jlocoPrivate->speedRef.w);
    else
      fprintf(stderr, "  no ref\n");
    // speed line
    fprintf(stderr, "  vel (%5.2f   %5.2f) (%5.2f   %5.2f) (%5.2f  %5.2f)\n",
	    vRefRaw, wRefRaw, vRefLim, wRefLim, wLeft, wRight);
    prevmsg = msg;
#endif // STATEDBG
    
    PVDEBUG("raw(%f   %f) ==> lim(%f   %f) ==> act(%f  %f)\n",
	   vRefRaw, wRefRaw, vRefLim, wRefLim, wLeft, wRight);
    
    // Example gnuplot commands (each on ONE line):
    // plot "jloco.plot" u 0:1 w l title "raw v",
    //      "jloco.plot" u 0:3 w l title "actual v",
    //      "jloco.plot" u 0:5 w l title "limit v"
    // plot "jloco.plot" u 0:7 w l title "left wheel ref",
    //      "jloco.plot" u 0:9 w l title "left wheel actual"
    PPLOT("%g %g %g %g %g %g %g %g %g %g\n",
	  vRefRaw, wRefRaw, vCurrent, wCurrent, vRefLim, wRefLim,
	  wLeft, wRight,
	  jlocoPrivate->odo.leftWheel.w, jlocoPrivate->odo.rightWheel.w);

    prevV = vRefLim;
    prevW = wRefLim;
    
#if INVERT_LEFT_WHEEL == 1
    wLeft = - wLeft;
#elif INVERT_LEFT_WHEEL != 0
# error INVERT_LEFT_WHEEL must be 0 or 1
#endif
#if INVERT_RIGHT_WHEEL == 1
    wRight = - wRight;
#elif INVERT_RIGHT_WHEEL != 0
# error INVERT_RIGHT_WHEEL must be 0 or 1
#endif
    {
      const int res = transmitMotorSpeeds(neo, wLeft, wRight,
					  NEO_MOTORS_MAXNTRIALS);
      if(0 != res){
	PERRMSG("transmitMotorSpeeds() returned %d\n", res);
	*report = S_jloco_MOTOR_ERROR;
	return ERROR;
      }
    }
  }
  
  return OK;
}


/*------------------------------------------------------------------------
 * endOdoAndAsserv  -  Termination codel of CmdTask
 * 
 * Returns:    OK or ERROR
 */
STATUS
endOdoAndAsserv(void)
{
  if(NULL == neo){
    PDEBUG("neoEnd() already done\n");
    return OK;
  }
  PDEBUG("calling neoEnd()\n");
  neoEnd(neo);
  neo = NULL;
  return OK;
}

/*----------------------------------------------------------------------
 * SetPom
 *
 * Description:
 *
 * Reports:	OK
 *		S_jloco_POSTER_NOT_FOUND
 */

/* setPom - codel EXEC of SetPom
   Returns: EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
setPom(JLOCO_NAME * posterName, int * report)
{
  POSTER_ID id;
  if(OK != posterFind(posterName->str, &id)){ // posterFind() should use const
    PDEBUG("posterFind(%s) failed\n", posterName->str);
    *report = S_jloco_POSTER_NOT_FOUND;
  }
  else
    pomPosterId = id;
  return ETHER;
}

/*------------------------------------------------------------------------
 * JoystickOn
 *
 * Description: 
 *
 * Reports:      OK
 *              S_jloco_NEOBOTIX_ERROR
 *              S_jloco_MOTOR_ERROR
 *              S_jloco_JOYSTICK_ERROR
 *              S_jloco_MOTOR_READ_ERROR
 *              S_jloco_INVALID_TRACK_MODE
 *              S_jloco_BATTERY_LOW
 *              S_jloco_SIGACTION_ERROR
 */

/* jlocoJoystickOnStart  -  codel START of JoystickOn
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
jlocoJoystickOnStart(int *report)
{
  jlocoPrivate->trackMode = JLOCO_MANUAL;
  return EXEC;
}

/* jlocoJoystickOnMain  -  codel EXEC of JoystickOn
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
jlocoJoystickOnMain(int *report)
{
  if (EXEC_TASK_BILAN(JLOCO_CMDTASK_NUM) != OK) {
    *report = EXEC_TASK_BILAN(JLOCO_CMDTASK_NUM);
    jlocoPrivate->trackMode = JLOCO_IDLE;
    return ETHER;
  }
  /* Check for user abort (button 4) */
  if (joy->buttons[3]) {
    jlocoPrivate->trackMode = JLOCO_IDLE;
    return ETHER;
  }
  return EXEC;
}

/* jlocoJoystickOnInter  -  codel INTER of JoystickOn
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
jlocoJoystickOnInter(int *report)
{
  jlocoPrivate->trackMode = JLOCO_IDLE;
  return ETHER;
}



/*----------------------------------------------------------------------*/
/*
 * begin
 * static function implementations
 */

static int updateSpeedCache(JLOCO_ODO *odo, double motRr, double motRl,
		     double motW)
{
  double dpr, dpl, velr, vell, vCurrent, wCurrent;
  const int res =
    receiveMotorChange(neo, &dpl, &vell, &dpr, &velr, NEO_MOTORS_MAXNTRIALS);
  if(0 != res){
    PDEBUG("receiveMotorChange() returned %d\n", res);
    return res;
  }
  
#if INVERT_LEFT_WHEEL == 1
  dpl = - dpl;
  vell = - vell;
#elif INVERT_LEFT_WHEEL != 0
# error INVERT_LEFT_WHEEL must be 0 or 1
#endif
#if INVERT_RIGHT_WHEEL == 1
  dpr = - dpr;
  velr = - velr;
#elif INVERT_RIGHT_WHEEL != 0
# error INVERT_RIGHT_WHEEL must be 0 or 1
#endif
  
  odo->rightWheel.phi += dpr;
  odo->rightWheel.w    = velr;
  odo->rightWheel.v    = motRr * velr;
  odo->rightWheel.dPhi = dpr;
  odo->rightWheel.dS   = motRr * dpr;
  
  odo->leftWheel.phi += dpl;
  odo->leftWheel.w    = vell;
  odo->leftWheel.v    = motRl * vell;
  odo->leftWheel.dPhi = dpl;
  odo->leftWheel.dS   = motRl * dpl;
  
  speedWheelsToGlob(odo->rightWheel.w, odo->leftWheel.w,
		    &vCurrent, &wCurrent,
		    motRr, motRl, motW);
  odo->robot.v = vCurrent;
  odo->robot.w = wCurrent;
  
  return 0;
}


static void speedGlobToWheels(double v, double w,
		       double *wRight, double *wLeft,
		       double motRr, double motRl, double motW)
{
  if(fabs(v) < JLOCO_EPSILON && fabs(w) < JLOCO_EPSILON){
    *wRight = 0;
    *wLeft = 0;
  }
  else{
    const double deltaV = motW * w / 2.;
    *wRight = (v + deltaV) / motRr;
    *wLeft = (v - deltaV) / motRl;
  }
}


void speedWheelsToGlob(double wRight, double wLeft, double *v, double *w,
		       double motRr, double motRl, double motW)
{
  if(fabs(wRight) < JLOCO_EPSILON && fabs(wLeft) < JLOCO_EPSILON){
    *v = 0;
    *w = 0;
  }
  else{
    wRight *= motRr;
    wLeft  *= motRl;
    *v = 0.5 * (wLeft + wRight);
    *w = (wRight - wLeft) / motW;
  }
}


static double applyLimits(double current, double wanted,
		   double limit, double acceleration,
		   double timeslice)
{
  const double maxDelta = timeslice * acceleration;
  double delta;
  
  // speed limit:
  if(wanted >   limit) wanted =   limit;
  if(wanted < - limit) wanted = - limit;
  
  // acceleration limit:
  delta = wanted - current;
  if(delta >   maxDelta) return current + maxDelta;
  if(delta < - maxDelta) return current - maxDelta;
  
  return wanted;
}


static void getJoystickSpeeds(double *v, double *w)
{
  if (joy->buttons[0]) {
    *v = joy->axes[1] * JLOCO_JOY_VGAIN;
    *w = joy->axes[0] * JLOCO_JOY_WGAIN;
  }
  else {
    *v = 0;
    *w = 0;
  }
}


static int checkRobotMotion(double vWanted, double wWanted,
		     double vCurrent, double wCurrent)
{
  if((fabs(vWanted) > 10 * JLOCO_EPSILON)
     || (fabs(wWanted) > 10 * JLOCO_EPSILON)) {
    if (fabs(vWanted - vCurrent) > 0.05 ||
	fabs(wWanted - wCurrent) > 0.05) {
      // translational or rotational speed mismatch
      PDEBUG("MISMATCH wanted=(%f   %f)  actual(%f   %f)\n",
	     vWanted, wWanted,  vCurrent, wCurrent);
      return 1;
    }
  }
  return 0;
}


static void updateOdometry(JLOCO_ODO *odo, double delta_gyro_angle, double motW,
		    double qCoeff)
{
  if(JLOCO_GYRO_MODE_ENABLE == SDI_F->gyroMode){
    // Hacked adaptation to braindead legacy API.  Note that
    // odoGyroProba() does not corectly propagate the covariances
    // anyways and uses a simplistic "half angle straight line" model
    // instead of a curvilinear one.
    JLOCO_GYRO gyro;
    memcpy(&gyro.robot,     &odo->robot,     sizeof(gyro.robot));
    memcpy(&gyro.odoError,  &odo->odoError,  sizeof(gyro.odoError));
    memcpy(&gyro.robotMisc, &odo->robotMisc, sizeof(gyro.robotMisc));
    gyro.robotMisc.dTheta = delta_gyro_angle;
    gyro.robot.w = delta_gyro_angle / jlocoPrivate->cmdTaskPeriod;
    odoProba0(odo, motW, qCoeff);
    odoGyroProba(odo, &gyro, motW);
    memcpy(&odo->robot,     &gyro.robot,     sizeof(odo->robot));
    memcpy(&odo->odoError,  &gyro.odoError,  sizeof(odo->odoError));
    memcpy(&odo->robotMisc, &gyro.robotMisc, sizeof(odo->robotMisc));
  }
  else
    odoProba0(odo, motW, qCoeff);
}


static void updatePomPosters(JLOCO_PRIVATE *jp)
{
  // NOTE: see the (French) comments in calife/src/matrice/matrice.c
  // for documentation about how to get the co-variances out of the
  // odoError array:
  // the matrix [ 0 1 3 ]
  //            [ 1 2 4 ]  
  //            [ 3 4 5 ]
  // is stored as the array [0 1 2 3 4 5].
  
  POM_ME_POS *mePos = &SDI_F->mePos;
  if(NULL != pomPosterId){
    POM_POS pomPos;
    if(OK == jlocoPOM_POSPosterRead(pomPosterId, &pomPos)) {
      mePos->date1 = pomPos.date;
    }
    else{
      PDEBUG_ERR("jlocoPOM_POSPosterRead() failed\n");
      pomPosterId = NULL;
      mePos->date1 = -1;
    }
  }
  else
      mePos->date1 = -1;
  mePos->main.euler.yaw = jp->odo.robot.theta;
  // pitch and roll remain zero
  mePos->main.euler.x   = jp->odo.robot.xRob;
  mePos->main.euler.y   = jp->odo.robot.yRob;
  // z remains zero
  mePos->main.var.voo   = jp->odo.odoError.var[5];
  // vpp and vrr remain JLOCO_VAR_TT
  mePos->main.var.vxx   = jp->odo.odoError.var[0];
  mePos->main.var.vyy   = jp->odo.odoError.var[2];
  // vzz remains JLOCO_VAR_XX
  mePos->main.var.vox   = jp->odo.odoError.var[3];
  mePos->main.var.voy   = jp->odo.odoError.var[4];
  mePos->main.var.vxy   = jp->odo.odoError.var[1];
  // all other co-variances remain zero
  
  // speeds are mostly zero, just update the linear and angular components
  mePos->VLocal.euler.x   = jp->odo.robot.v;
  mePos->VLocal.euler.yaw = jp->odo.robot.w;
  // keep speed variances zero
}


static void initCmdParam(GENPOS_CART_CMD_PARAM *cmd)
{
  cmd->distPoint = JLOCO_CNTR_POINT_DIST;
  cmd->KpLongit  = JLOCO_CNTR_KP_LONGIT;
  cmd->KiLongit  = JLOCO_CNTR_KI_LONGIT;
  cmd->KpTransv  = JLOCO_CNTR_KP_TRANSV;
  cmd->KiTransv  = JLOCO_CNTR_KI_TRANSV;
  cmd->KpAng     = JLOCO_CNTR_KP_ANG;
  cmd->KiAng     = JLOCO_CNTR_KI_ANG;
  cmd->longitErrorMax  = JLOCO_CNTR_LONGIT_ERROR_MAX;
  cmd->transvErrorMax  = JLOCO_CNTR_TRANSV_ERROR_MAX;
  cmd->longitMaxErrSum = JLOCO_CNTR_LONGIT_MAX_ERR_SUM;
  cmd->transvMaxErrSum = JLOCO_CNTR_TRANSV_MAX_ERR_SUM;
  cmd->angErrorMax     = JLOCO_CNTR_ANG_ERROR_MAX;
  cmd->angMaxErrSum    = JLOCO_CNTR_ANG_MAX_ERR_SUM;
  cmd->longitMaxGap    = JLOCO_CNTR_LONGIT_GAP_MAX;
  cmd->transvMaxGap    = JLOCO_CNTR_TRANSV_GAP_MAX;
  cmd->angMaxGap       = JLOCO_CNTR_ANG_GAP_MAX;
}


static void initVarParams(JLOCO_VAR_PARAMS *vp)
{
  vp->vxx    = JLOCO_VAR_XX;
  vp->vyy    = JLOCO_VAR_YY;
  vp->vtt    = JLOCO_VAR_TT;
  vp->qCoeff = JLOCO_VAR_QCOEFF;
}


static void initOdoModel(JLOCO_PRIVATE *jp)
{
  SDI_F->tmpOdoCorr.right_mult = JLOCO_RIGHT_CORRECTION;
  SDI_F->tmpOdoCorr.left_mult = JLOCO_LEFT_CORRECTION;
  SDI_F->tmpOdoCorr.axis_mult = JLOCO_AXIS_CORRECTION;

  jp->motRr_nom = JLOCO_RIGHT_WHEEL_RADIUS;
  jp->motRl_nom = JLOCO_LEFT_WHEEL_RADIUS;
  jp->motW_nom  = JLOCO_AXIS_WIDTH;

  jp->motRr_cor = JLOCO_RIGHT_WHEEL_RADIUS * SDI_F->tmpOdoCorr.right_mult;
  jp->motRl_cor = JLOCO_LEFT_WHEEL_RADIUS  * SDI_F->tmpOdoCorr.left_mult;
  jp->motW_cor  = JLOCO_AXIS_WIDTH         * SDI_F->tmpOdoCorr.axis_mult;
}


static void initLimits(JLOCO_PRIVATE *jp)
{
  jp->sdMax      = JLOCO_VMAX_ROBOT;
  jp->thetadMax  = JLOCO_WMAX_ROBOT;
  jp->sddMax     = JLOCO_AMAX_LIN_ROBOT;
  jp->thetaddMax = JLOCO_AMAX_ANG_ROBOT;
}


static void initWatchdogs(JLOCO_PRIVATE *jp)
{
  jp->waitRefCounter        = 0;
  jp->cmdMismatchThreshold  = JLOCO_CMD_MISMATCH_THRESHOLD;
  jp->pumpMismatchThreshold = JLOCO_PUMP_MISMATCH_THRESHOLD;
  jp->importantDrift        = 0;
  SDI_F->watchdogTolerance  = JLOCO_DEFAULT_WATCHDOG_TOLERANCE;
}


static void initPomPosters()
{
  POM_ME_POS *mePos = &SDI_F->mePos;
  mePos->kind = POM_ME_DELTA;
  mePos->date1 = -1;
  mePos->confidence = 0.2;	// XXX used?
  mePos->main.var.voo     = JLOCO_VAR_TT;
  mePos->main.var.vpp     = JLOCO_VAR_TT;
  mePos->main.var.vrr     = JLOCO_VAR_TT;
  mePos->main.var.vxx     = JLOCO_VAR_XX;
  mePos->main.var.vyy     = JLOCO_VAR_YY;
  mePos->main.var.vzz     = JLOCO_VAR_XX;
}


static int keepMotorsAlive(NEO_STR *neo, int maxtrials)
{
  CAN_MOTOR_STATE ls, rs;
  ls = canMotorsState(neo, CAN_MOTOR_LEFT);
  rs = canMotorsState(neo, CAN_MOTOR_RIGHT);
  if((ST_MOTOR_FAILURE == ls) || (ST_MOTOR_FAILURE == rs))
    return -1;
  while((ST_OPERATION_ENABLED != ls) || (ST_OPERATION_ENABLED != rs)){
    --maxtrials;
    if(0 > maxtrials)
      return -2;
    if(OK != canMotorsReset(neo))
      return -3;
    ls = canMotorsState(neo, CAN_MOTOR_LEFT);
    rs = canMotorsState(neo, CAN_MOTOR_RIGHT);
  }
  return 0;
}


static int transmitMotorSpeeds(NEO_STR *neo, double lvel, double rvel, int maxtrials)
{
  const int res = keepMotorsAlive(neo, maxtrials);
  if(0 != res)
    return res;
  canMotorsSetVelGearRadS(neo, lvel, rvel);
  return 0;
}


static int receiveMotorChange(NEO_STR *neo, double *lDeltaPos, double *lvel,
		       double *rDeltaPos, double *rvel, int maxtrials)
{
  const int res = keepMotorsAlive(neo, maxtrials);
  if(0 != res)
    return res;
  canMotorsGetGearDeltaPosVelRadS(neo, lDeltaPos, lvel, rDeltaPos, rvel);
  return 0;
}
